/*-
 * Id: if_fet.c, 2000/10/30 AJ Jiang & Saten Shih & Yiping Chen, VIA Technologies, Inc.
 *		ver 1.19
 *
 */

#include "bpfilter.h"

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/sockio.h>
#include <sys/mbuf.h>
#include <sys/malloc.h>
#include <sys/kernel.h>
#include <sys/socket.h>

#include <net/if.h>
#include <net/if_arp.h>
#include <net/ethernet.h>
#include <net/if_dl.h>
#include <net/if_media.h>

#if NBPFILTER > 0
#include <net/bpf.h>
#endif

#include <vm/vm.h>              /* for vtophys */
#include <vm/pmap.h>            /* for vtophys */
#include <machine/clock.h>      /* for DELAY */
#include <machine/bus_pio.h>
#include <machine/bus_memio.h>
#include <machine/bus.h>

#include <pci/pcireg.h>
#include <pci/pcivar.h>

#define FET_USEIOSPACE
/*
#define DEBUG
*/

/* #define FET_BACKGROUND_AUTONEG */

#include <pci/if_fet.h>


/*
 * Various supported device vendors/types and their names.
 */
static struct fet_type fet_devs[] = {
	{ VIA_VENDORID, VIA_DEVICEID_3043,
		"VIA VT86C100A 10/100Mb Fast Ethernet Adapter" },
	{ VIA_VENDORID, VIA_DEVICEID_3065,
		"VIA VT6102 PCI 10/100Mb Fast Ethernet Adapter" },
	{ 0, 0, NULL }
};


static unsigned long fet_count = 0;
static const char *fet_probe	__P((pcici_t, pcidi_t));
static void fet_attach		__P((pcici_t, int));

static int fet_newbuf		__P((struct fet_softc *,
						struct fet_chain_onefrag *));
static int fet_encap		__P((struct fet_softc *, struct fet_chain *,
						struct mbuf * ));

static void fet_rxeof		__P((struct fet_softc *));
static void fet_rxeoc		__P((struct fet_softc *));
static void fet_txeof		__P((struct fet_softc *));
static void fet_txeoc		__P((struct fet_softc *));
static void fet_intr		__P((void *));
static void fet_start		__P((struct ifnet *));
static int fet_ioctl		__P((struct ifnet *, u_long, caddr_t));
static void fet_init		__P((void *));
static void fet_stop		__P((struct fet_softc *));
static void fet_watchdog		__P((struct ifnet *));
static void fet_shutdown		__P((int, void *));
static int fet_ifmedia_upd	__P((struct ifnet *));
static void fet_ifmedia_sts	__P((struct ifnet *, struct ifmediareq *));
static void fet_link_change	__P((struct fet_softc *));

static unsigned fet_read_mii __P((struct fet_softc *, char));
static void fet_write_mii    __P((struct fet_softc *, char, char, char));
static void fet_mii_delay    __P((void));
static unsigned read_tick_counter	__P((void));
static int fet_query_auto    __P((struct fet_softc *));
static void post_reset_delay	__P((void));

static u_int8_t fet_calchash	__P((u_int8_t *));
static void fet_setmulti		__P((struct fet_softc *));
static void fet_reset		__P((struct fet_softc *));
static int fet_list_rx_init	__P((struct fet_softc *));
static int fet_list_tx_init	__P((struct fet_softc *));
static int fet_safe_rx_off	__P((struct fet_softc *));

#define FET_SETBIT(sc, reg, x)				\
	CSR_WRITE_1(sc, reg,				\
		CSR_READ_1(sc, reg) | x)

#define FET_CLRBIT(sc, reg, x)				\
	CSR_WRITE_1(sc, reg,				\
		CSR_READ_1(sc, reg) & ~x)

#define FET_SETBIT16(sc, reg, x)				\
	CSR_WRITE_2(sc, reg,				\
		CSR_READ_2(sc, reg) | x)

#define FET_CLRBIT16(sc, reg, x)				\
	CSR_WRITE_2(sc, reg,				\
		CSR_READ_2(sc, reg) & ~x)

#define FET_SETBIT32(sc, reg, x)				\
	CSR_WRITE_4(sc, reg,				\
		CSR_READ_4(sc, reg) | x)

#define FET_CLRBIT32(sc, reg, x)				\
	CSR_WRITE_4(sc, reg,				\
		CSR_READ_4(sc, reg) & ~x)

#define SIO_SET(x)					\
	CSR_WRITE_1(sc, FET_MIICR,			\
		CSR_READ_1(sc, FET_MIICR) | x)

#define SIO_CLR(x)					\
	CSR_WRITE_1(sc, FET_MIICR,			\
		CSR_READ_1(sc, FET_MIICR) & ~x)

static int fet_safe_rx_off(sc)
	struct fet_softc		*sc;
{
	u_int16_t	ww;
	
	FET_CLRBIT16(sc, FET_COMMAND, FET_CMD_RX_ON);
	post_reset_delay();
	
	for (ww = 0; ww < W_MAX_TIMEOUT; ww++)
		if (!(CSR_READ_2(sc, FET_COMMAND) & FET_CMD_RX_ON))
			break;

	if (ww == W_MAX_TIMEOUT) {
		if(sc->fet_chip_revid < REV_ID_VT3065_A)
			return FALSE;
		else {
			/* set MAC to internal loopback */
			FET_SETBIT(sc, FET_TXCFG, FET_TXCFG_LB0);

			/* enter FIFO test mode to issue RX reject */
			FET_SETBIT(sc, FET_GFTEST, 0x01);
			FET_SETBIT(sc, FET_RFTCMD, 0x08);	
			post_reset_delay();
			FET_SETBIT(sc, FET_GFTEST, 0x00);

			/* set MAC to normal */
			FET_CLRBIT(sc, FET_TXCFG, FET_TXCFG_LB0);

			return TRUE;
			}		
		}

	else return TRUE;
}

static unsigned fet_read_mii(sc, byMIIIndex)
	struct fet_softc		*sc;
	char	byMIIIndex;
{
	unsigned ReturnMII;
	char byMIIAdrbak;
	char byMIICRbak;
	char byMIItemp;
	int s;

	s = splimp();

	byMIIAdrbak = CSR_READ_1(sc, FET_MIIADDR);
	byMIICRbak = CSR_READ_1(sc, FET_MIICR);

	CSR_WRITE_1(sc, FET_MIICR, byMIICRbak & ~FET_MIICR_MAUTO);    

	if (sc->fet_chip_revid < REV_ID_VT3071_A)
		fet_mii_delay();
	else
		while(!(CSR_READ_1(sc, FET_MIIADDR) & FET_MIIADDR_MIDLE));

	CSR_WRITE_1(sc, FET_MIIADDR, byMIIIndex);
	fet_mii_delay();

	FET_SETBIT(sc, FET_MIICR, FET_MIICR_RCMD);

	byMIItemp = CSR_READ_1(sc, FET_MIICR);
	byMIItemp = byMIItemp & FET_MIICR_RCMD;

	while (byMIItemp != 0) {
		   byMIItemp = CSR_READ_1(sc, FET_MIICR);
		   byMIItemp = byMIItemp & FET_MIICR_RCMD;
		  }
	fet_mii_delay();

	ReturnMII = CSR_READ_2(sc, FET_MIIDATA);

	CSR_WRITE_1(sc, FET_MIIADDR, byMIIAdrbak);
	CSR_WRITE_1(sc, FET_MIICR, byMIICRbak);
	fet_mii_delay();

	(void)splx(s);
	return(ReturnMII);
}

static void fet_write_mii(sc, byMIISetByte, byMIISetBit, byMIIOP)
	struct fet_softc		*sc;
	char	byMIISetByte;
	char	byMIISetBit;
	char	byMIIOP;
{
	unsigned ReadMIItmp;
	unsigned MIIMask;
	char byMIIAdrbak;
	char byMIICRbak;
	char byMIItemp;
	int  s;

	s = splimp();

	byMIIAdrbak = CSR_READ_1(sc, FET_MIIADDR);
	byMIICRbak= CSR_READ_1(sc, FET_MIICR);

	CSR_WRITE_1(sc, FET_MIICR, byMIICRbak & ~FET_MIICR_MAUTO);

	if(sc->fet_chip_revid < REV_ID_VT3071_A)
	   fet_mii_delay();
	else
	   while(!(CSR_READ_1(sc, FET_MIIADDR) & FET_MIIADDR_MIDLE));

	CSR_WRITE_1(sc, FET_MIIADDR, byMIISetByte);
	fet_mii_delay();

	FET_SETBIT(sc, FET_MIICR, FET_MIICR_RCMD);

	byMIItemp = CSR_READ_1(sc, FET_MIICR);
	byMIItemp = byMIItemp & FET_MIICR_RCMD;

	while (byMIItemp != 0) {
		   byMIItemp = CSR_READ_1(sc, FET_MIICR);
		   byMIItemp = byMIItemp & FET_MIICR_RCMD;
		  }
	fet_mii_delay();

	ReadMIItmp = CSR_READ_2(sc, FET_MIIDATA);
	MIIMask = 0x0001;
	MIIMask = MIIMask << byMIISetBit;


	if (byMIIOP == 0) {
		MIIMask = ~MIIMask;
		ReadMIItmp = ReadMIItmp & MIIMask;
	   }
	else {
		ReadMIItmp = ReadMIItmp | MIIMask;
	   }

	CSR_WRITE_2(sc, FET_MIIDATA, ReadMIItmp);
	fet_mii_delay();

	FET_SETBIT(sc, FET_MIICR, FET_MIICR_WCMD);
	byMIItemp = CSR_READ_1(sc, FET_MIICR);
	byMIItemp = byMIItemp & FET_MIICR_WCMD;

	while (byMIItemp != 0) {
		   byMIItemp = CSR_READ_1(sc, FET_MIICR);
		   byMIItemp = byMIItemp & FET_MIICR_WCMD;
		  }
	fet_mii_delay();

	CSR_WRITE_1(sc, FET_MIIADDR, byMIIAdrbak & ~FET_MIICR_MAUTO);
	CSR_WRITE_1(sc, FET_MIICR, byMIICRbak);
	fet_mii_delay();

	(void)splx(s);
}

void fet_mii_delay(void)
{
	u_int16_t i;

	for	(i=0;i<0x7fff;i++){
		inb(0x61);
		inb(0x61);
		inb(0x61);
		inb(0x61);
	  }
}

void post_reset_delay(void)
{

	unsigned initial_time, current_time;
	
	/* need to wait 10 mini seconds */

	initial_time = read_tick_counter();

	/* 8254 time tick gives 419 nsec per count. We have to delay a delta
		count of 10,000,000ns/419ns = 23866. We will round it up to 23866
	*/

	while (1){
		current_time = read_tick_counter();
		if ((initial_time - current_time) >= 23866)		/* 10 mini second */
			{
			return;
			}
		}
}

unsigned read_tick_counter(void)
{
	unsigned current_time = 0;

	/* Program 8254 to latch T0's count */
	outb(0x6, 0x43);
	/* Read the counter */
	current_time = inb(0x40);
	current_time = current_time | (inb(0x40) << 8);
	return current_time;
}

/*
 * Query duplex mode
*/

static int fet_query_auto(sc)
	struct fet_softc		*sc;
{
	u_int16_t FDXFlag, MIIReturn;

	MIIReturn = fet_read_mii(sc, PHY_ANEXP);

	if ((MIIReturn & PHY_ANEXP_LPAUTOABLE) == 0) {
		 FDXFlag = 0;
		 return FDXFlag;
	   }

	MIIReturn = fet_read_mii(sc, PHY_LPAR);

	if ((MIIReturn & (PHY_LPAR_100BTXFULL | PHY_LPAR_10BTFULL)) == 0) {
		 FDXFlag = 0;
		 return FDXFlag;
	   }

	MIIReturn = fet_read_mii(sc, PHY_ANAR);

	if ((MIIReturn & (PHY_ANAR_100BTXFULL | PHY_ANAR_10BTFULL)) == 0) {
		 FDXFlag = 0;
		 return FDXFlag;
	   }

	FDXFlag = 1;
	return FDXFlag;
}

/* 
 * process link status change in ISR
 */
static void fet_link_change(sc)
	struct fet_softc		*sc;
{
	u_int8_t FDXFlag;

	if (CSR_READ_1(sc, FET_MIISR) & FET_MIISR_LNKFL) {		
		if (sc->fet_modeinfo == FET_AUTO) {
			if (sc->fet_chip_revid < REV_ID_VT3071_A)				
				fet_write_mii(sc, PHY_BMCR, PHY_BMCR_RESET_OFFSET, 1);
		}
	}
	else {		
		if (sc->fet_modeinfo == FET_AUTO) {

			/* special treatment for VT3071B's LED */
              if(sc->fet_chip_revid == REV_ID_VT3071_B) {
				if (CSR_READ_1(sc, FET_MIISR) & FET_MIISR_SPEED)
					FET_CLRBIT(sc, FET_BCR0, FET_BCR0_LED100M);
				else
					FET_SETBIT(sc, FET_BCR0, FET_BCR0_LED100M);
				}

			FDXFlag = fet_query_auto(sc);

			if (FDXFlag == 1)		 	
				FET_SETBIT16(sc, FET_COMMAND, FET_CMD_FDX);       /* Set Chip Fullduplex mode */
			else
				FET_CLRBIT16(sc, FET_COMMAND, FET_CMD_FDX);      /* Set Chip Halfduplex mode */
			}
		}
}

		             
/*
 * Calculate CRC of a multicast group address, return the lower 6 bits.
 */
static u_int8_t fet_calchash(addr)
	u_int8_t		*addr;
{
	u_int32_t		crc, carry;
	int			i, j;
	u_int8_t		c;

	/* Compute CRC for the address value. */
	crc = 0xFFFFFFFF; /* initial value */

	for (i = 0; i < 6; i++) {
		c = *(addr + i);
		for (j = 0; j < 8; j++) {
			carry = ((crc & 0x80000000) ? 1 : 0) ^ (c & 0x01);
			crc <<= 1;
			c >>= 1;
			if (carry)
				crc = (crc ^ 0x04c11db6) | carry;
		}
	}

	/* return the filter bit position */
	return((crc >> 26) & 0x0000003F);
}

/*
 * Program the 64-bit multicast hash filter.
 */
static void fet_setmulti(sc)
	struct fet_softc		*sc;
{
	struct ifnet		*ifp;
	int			h = 0;
	u_int32_t		hashes[2] = { 0, 0 };
	struct ifmultiaddr	*ifma;
	u_int8_t		rxfilt;
	int			mcnt = 0;

	ifp = &sc->arpcom.ac_if;

	rxfilt = CSR_READ_1(sc, FET_RXCFG);

	if (ifp->if_flags & IFF_ALLMULTI || ifp->if_flags & IFF_PROMISC) {
		rxfilt |= FET_RXCFG_RX_MULTI;
		CSR_WRITE_1(sc, FET_RXCFG, rxfilt);
		CSR_WRITE_4(sc, FET_MAR0, 0xFFFFFFFF);
		CSR_WRITE_4(sc, FET_MAR1, 0xFFFFFFFF);
		return;
	}

	/* first, zot all the existing hash bits */
	CSR_WRITE_4(sc, FET_MAR0, 0);
	CSR_WRITE_4(sc, FET_MAR1, 0);

	/* now program new ones */
	for (ifma = ifp->if_multiaddrs.lh_first; ifma != NULL;
				ifma = ifma->ifma_link.le_next) {
		if (ifma->ifma_addr->sa_family != AF_LINK)
			continue;
		h = fet_calchash(LLADDR((struct sockaddr_dl *)ifma->ifma_addr));
		if (h < 32)
			hashes[0] |= (1 << h);
		else
			hashes[1] |= (1 << (h - 32));
		mcnt++;
	}

	if (mcnt)
		rxfilt |= FET_RXCFG_RX_MULTI;
	else
		rxfilt &= ~FET_RXCFG_RX_MULTI;

	CSR_WRITE_4(sc, FET_MAR0, hashes[0]);
	CSR_WRITE_4(sc, FET_MAR1, hashes[1]);
	CSR_WRITE_1(sc, FET_RXCFG, rxfilt);

	return;
}

static void fet_reset(sc)
	struct fet_softc		*sc;
{	
	u_int16_t	ww;

	FET_SETBIT16(sc, FET_COMMAND, FET_CMD_RESET);

	for (ww = 0; ww < W_MAX_TIMEOUT; ww++) {
		DELAY(10);
		if (!(CSR_READ_2(sc, FET_COMMAND) & FET_CMD_RESET))
			break;
	}
	if (ww == W_MAX_TIMEOUT) {
		if (sc->fet_chip_revid < REV_ID_VT3065_A)
			printf("fet%d: reset never completed!\n", sc->fet_unit);
		else
			/* turn on force reset */
			FET_SETBIT(sc, FET_MISC_CR1, FET_MISCCR1_FORSRST);
		}
			

	/* Wait a little while for the chip to get its brains in order. */
	DELAY(1000);	
			
	return;
}

/*
 * Probe for a VIA Rhine chip. Check the PCI vendor and device
 * IDs against our list and return a device name if we find a match.
 */
static const char *
fet_probe(config_id, device_id)
	pcici_t			config_id;
	pcidi_t			device_id;
{
	struct fet_type		*t;

	t = fet_devs;

	while(t->fet_name != NULL) {
		if ((device_id & 0xFFFF) == t->fet_vid &&
		    ((device_id >> 16) & 0xFFFF) == t->fet_did) {
			return(t->fet_name);
		}
		t++;
	}

	return(NULL);
}

/*
 * Attach the interface. Allocate softc structures, do ifmedia
 * setup and ethernet/BPF attach.
 */
static void
fet_attach(config_id, unit)
	pcici_t			config_id;
	int			unit;
{
	int			s, i;
#ifndef FET_USEIOSPACE
	vm_offset_t		pbase, vbase;
#endif
	u_char			eaddr[ETHER_ADDR_LEN];
	u_int32_t		command, id;
	struct fet_softc		*sc;
	struct ifnet		*ifp;
	unsigned int		round;
	caddr_t			roundptr;	
	u_int16_t		nmedia, defmedia;
	const int		*media;
	int 			LineSpeed, MIICRbak;
    unsigned char 	byBCR1Value;
	int 			FDXFlag, AssignFlag;
	int				idx, idx1;

	s = splimp();

	sc = malloc(sizeof(struct fet_softc), M_DEVBUF, M_NOWAIT);
	if (sc == NULL) {
		printf("fet%d: no memory for softc struct!\n", unit);
		return;
	}
	bzero(sc, sizeof(struct fet_softc));

	/*
	 * Handle power management nonsense.
	 */

	command = pci_conf_read(config_id, FET_PCI_CAPID) & 0x000000FF;
	if (command == 0x01) {

		command = pci_conf_read(config_id, FET_PCI_PWRMGMTCTRL);
		if (command & FET_PSTATE_MASK) {
			u_int32_t		iobase, membase, irq;

			/* Save important PCI config data. */
			iobase = pci_conf_read(config_id, FET_PCI_LOIO);
			membase = pci_conf_read(config_id, FET_PCI_LOMEM);
			irq = pci_conf_read(config_id, FET_PCI_INTLINE);

			/* Reset the power state. */
			printf("fet%d: chip is in D%d power mode "
			"-- setting to D0\n", unit, command & FET_PSTATE_MASK);
			command &= 0xFFFFFFFC;
			pci_conf_write(config_id, FET_PCI_PWRMGMTCTRL, command);

			/* Restore PCI config data. */
			pci_conf_write(config_id, FET_PCI_LOIO, iobase);
			pci_conf_write(config_id, FET_PCI_LOMEM, membase);
			pci_conf_write(config_id, FET_PCI_INTLINE, irq);
		}
	}

	/*
	 * Map control/status registers.
	 */
	command = pci_conf_read(config_id, PCI_COMMAND_STATUS_REG);
	command |= (PCIM_CMD_PORTEN|PCIM_CMD_MEMEN|PCIM_CMD_BUSMASTEREN);
	pci_conf_write(config_id, PCI_COMMAND_STATUS_REG, command);
	command = pci_conf_read(config_id, PCI_COMMAND_STATUS_REG);

	sc->fet_chip_revid  = pci_conf_read(config_id, FET_PCI_REVID) & 0x000000FF;
	id = pci_conf_read(config_id, FET_PCI_VENDOR_ID);

	if (PCI_VENDORID(id) != VIA_VENDORID) {
		printf("fet%d: wrong vendor ID !\n", unit);
		goto fail;
	}

	if (!(PCI_CHIPID(id) == VIA_DEVICEID_3043 ||
				 PCI_CHIPID(id) == VIA_DEVICEID_3065)) {
		printf("fet%d: wrong device ID !\n", unit);
		goto fail;
	}

#ifdef FET_USEIOSPACE
	if (!(command & PCIM_CMD_PORTEN)) {
		printf("fet%d: failed to enable I/O ports!\n", unit);
		free(sc, M_DEVBUF);
		goto fail;
	}

	if (!pci_map_port(config_id, FET_PCI_LOIO,
					(u_int16_t *)(&sc->fet_bhandle))) {
		printf ("fet%d: couldn't map ports\n", unit);
		goto fail;
	}
	sc->fet_btag = I386_BUS_SPACE_IO;
#else
	if (!(command & PCIM_CMD_MEMEN)) {
		printf("fet%d: failed to enable memory mapping!\n", unit);
		goto fail;
	}

	if (!pci_map_mem(config_id, FET_PCI_LOMEM, &vbase, &pbase)) {
		printf ("fet%d: couldn't map memory\n", unit);
		goto fail;
	}

	sc->fet_bhandle = vbase;
	sc->fet_btag = I386_BUS_SPACE_MEM;
#endif

	/* Allocate interrupt */
	if (!pci_map_int(config_id, fet_intr, sc, &net_imask)) {
		printf("fet%d: couldn't map interrupt\n", unit);
		goto fail;
	}

	/* Reset the adapter. */
	fet_reset(sc);

        /* Turn on bit2 (MIION) in PCI configuration register 0x53 during initialization */
        pci_conf_write(config_id, FET_PCI_MODE, pci_conf_read(config_id,FET_PCI_MODE)|(FET_MODE3_MIION<<24));
	/*
	 * Get station address. The way the Rhine chips work,
	 * you're not allowed to directly access the EEPROM once
	 * they've been programmed a special way. Consequently,
	 * we need to read the node address from the PAR0 and PAR1
	 * registers.
	 */
	FET_SETBIT(sc, FET_EECSR, FET_EECSR_LOAD);
	DELAY(200);
	for (i = 0; i < ETHER_ADDR_LEN; i++)
		eaddr[i] = CSR_READ_1(sc, FET_PAR0 + i);

	/*
	 * A Rhine chip was detected. Inform the world.
	 */
	printf("fet%d: Ethernet address: %6D\n", unit, eaddr, ":");

	/* print driver version */
	printf("fet%d: if_fet.c: v1.19 10/30/2000 \n", unit);

	sc->fet_unit = unit;
	bcopy(eaddr, (char *)&sc->arpcom.ac_enaddr, ETHER_ADDR_LEN);

	sc->fet_ldata_ptr = malloc(sizeof(struct fet_list_data) + 8,
				M_DEVBUF, M_NOWAIT);
	if (sc->fet_ldata_ptr == NULL) {
		free(sc, M_DEVBUF);
		printf("fet%d: no memory for list buffers!\n", unit);
		return;
	}

	sc->fet_ldata = (struct fet_list_data *)sc->fet_ldata_ptr;
	round = (unsigned int)sc->fet_ldata_ptr & 0xF;
	roundptr = sc->fet_ldata_ptr;
	for (i = 0; i < 8; i++) {
		if (round % 8) {
			round++;
			roundptr++;
		} else
			break;
	}
	sc->fet_ldata = (struct fet_list_data *)roundptr;
	bzero(sc->fet_ldata, sizeof(struct fet_list_data));

	ifp = &sc->arpcom.ac_if;
	ifp->if_softc = sc;
	ifp->if_unit = unit;
	ifp->if_name = "fet";
	ifp->if_mtu = ETHERMTU;
	ifp->if_flags = IFF_BROADCAST | IFF_SIMPLEX | IFF_MULTICAST;
	ifp->if_ioctl = fet_ioctl;
	ifp->if_output = ether_output;
	ifp->if_start = fet_start;
	ifp->if_watchdog = fet_watchdog;
	ifp->if_init = fet_init;
	ifp->if_baudrate = 10000000;
	ifp->if_snd.ifq_maxlen = FET_TX_LIST_CNT - 1;

	/*
	 * check if read the mode from EEPROM
	*/

	AssignFlag = 0;		/* initial value is 0 */

	if(CSR_READ_1(sc, FET_CFGC) & FET_CFGC_MEDEN){

		if(CSR_READ_1(sc, FET_BCR0) & FET_BCR0_MED2){
			byBCR1Value = CSR_READ_1(sc, FET_BCR1) & (FET_BCR1_MED0 | FET_BCR1_MED1);

		   if(byBCR1Value == 0x00){
				AssignFlag = 0;
				sc->fet_modeinfo = FET_AUTO;
			}
		}

		else{
			AssignFlag = 1;
			byBCR1Value = CSR_READ_1(sc, FET_BCR1) & (FET_BCR1_MED0 | FET_BCR1_MED1);

			if(byBCR1Value == (FET_BCR1_MED0 | FET_BCR1_MED1))
				sc->fet_modeinfo = FET_100_FDX;

			else if(byBCR1Value == FET_BCR1_MED1)
					sc->fet_modeinfo = FET_10_FDX;

			else if(byBCR1Value == FET_BCR1_MED0)
					sc->fet_modeinfo = FET_100_HDX;

			else if(byBCR1Value == 0x00)
					sc->fet_modeinfo = FET_10_HDX;
			}
		}

	if (AssignFlag){
		fet_write_mii(sc, PHY_BMCR, PHY_BMCR_AUTONEGENBL_OFFSET, 0); /* turn off N-way */
		printf("fet%d: ",unit);
		if (sc->fet_modeinfo == FET_100_FDX || sc->fet_modeinfo == FET_100_HDX){
			fet_write_mii(sc, PHY_BMCR, PHY_BMCR_SPEEDSEL_OFFSET, 1); /* Speed 100 */
			printf("LineSpeed=100Mb,");
		   }
		else{
			fet_write_mii(sc, PHY_BMCR, PHY_BMCR_SPEEDSEL_OFFSET, 1); /* Speed 10 */
			printf("LineSpeed=10Mb,");
			}

		if (sc->fet_modeinfo == FET_100_FDX || sc->fet_modeinfo == FET_10_FDX){
			fet_write_mii(sc, PHY_BMCR, PHY_BMCR_DUPLEX_OFFSET, 1);  /* write mii Fullduplex mode */
			FET_SETBIT16(sc, FET_COMMAND, FET_CMD_FDX);                /* Set Chip Fullduplex mode */
			printf("Fullduplex\n");
		   }
		else{
			fet_write_mii(sc, PHY_BMCR, PHY_BMCR_DUPLEX_OFFSET, 0); /* write miiHalfduplex mode */
			FET_CLRBIT16(sc, FET_COMMAND, FET_CMD_FDX);               /* Set Chip Fullduplex mode */
			printf("Halfduplex\n");
			}
	   }	/* end if (AssignFlag) */

	else {
		/* restart MII auto-negotiation*/
		fet_write_mii(sc, PHY_BMCR, PHY_BMCR_AUTONEGRSTR_OFFSET, 1);
		printf("Analyzing Media type of fet%d, this will take several seconds........",unit);
		for (idx=0;idx<5;idx++) {
			for (idx1=0;idx1<100;idx1++)
				post_reset_delay();
				if (fet_read_mii(sc, PHY_BMSR) & PHY_BMSR_AUTONEGCOMP)
					break;
			}
		printf("OK\n");

		/* query MII to know LineSpeed,duplex mode */
		LineSpeed = CSR_READ_1(sc, FET_MIISR) & FET_MIISR_SPEED;
		printf("fet%d: ",unit);

		if(LineSpeed==1)
			printf("Linespeed=10Mb,");
		else
			printf("Linespeed=100Mb,");

		if(sc->fet_chip_revid == REV_ID_VT3071_B){
			if (LineSpeed == 1)
				FET_CLRBIT(sc, FET_BCR0, FET_BCR0_LED100M);
			else
				FET_SETBIT(sc, FET_BCR0, FET_BCR0_LED100M);
			}

		FDXFlag = fet_query_auto(sc);
		if (FDXFlag == 1) {
		 	printf("Fullduplex\n");
		 	FET_SETBIT16(sc, FET_COMMAND, FET_CMD_FDX);       /* Set Chip Fullduplex mode */
		}
		else {
		  	printf("Halfduplex\n");
		  	FET_CLRBIT16(sc, FET_COMMAND, FET_CMD_FDX);      /* Set Chip Halfduplex mode */
		  	}
	  }
   
	/* special treatment on LED for various phys */
	if (sc->fet_chip_revid < REV_ID_VT3071_A)
		/* for NS or Mison phys, turn on bit 1 in register 17h, no effect to DAVICOM phy */
		fet_write_mii(sc,0x17,1,1);

	else if (sc->fet_chip_revid < REV_ID_VT3065_A)
			 /* for ESI phys, turn on bit 7 in register 17h */
			 fet_write_mii(sc,0x17,7,1);

	 /* turn on MII link change */
	MIICRbak = CSR_READ_1(sc, FET_MIICR);
	FET_CLRBIT(sc, FET_MIICR, FET_MIICR_MAUTO);
	fet_mii_delay();
	CSR_WRITE_1(sc, FET_MIIADDR, FET_MIIADDR_MSRCEN | FET_MIIADDR_MAD0);
	fet_mii_delay();
		  
	CSR_WRITE_1(sc, FET_MIICR, MIICRbak | FET_MIICR_MAUTO);
		
	/*
	 * Do ifmedia setup.
	 */

	media = fet_media_standard;
	nmedia = sizeof(fet_media_standard) / sizeof(fet_media_standard[0]);
	defmedia = (IFM_ETHER|IFM_AUTO);

	ifmedia_init(&sc->ifmedia, 0, fet_ifmedia_upd, fet_ifmedia_sts);

	for (idx = 0; idx < nmedia; idx++) {
		 ifmedia_add(&sc->ifmedia, media[idx], 0, NULL);
		}

	fet_init(sc);	
	fet_stop(sc);

	ifmedia_set(&sc->ifmedia, defmedia);
	
	/*
	 * Call MI attach routines.
	 */
	if_attach(ifp);
	ether_ifattach(ifp);

#if NBPFILTER > 0
	bpfattach(ifp, DLT_EN10MB, sizeof(struct ether_header));
#endif

	at_shutdown(fet_shutdown, sc, SHUTDOWN_POST_SYNC);

fail:
	splx(s);
	return;
}

/*
 * Initialize the transmit descriptors.
 */
static int fet_list_tx_init(sc)
	struct fet_softc		*sc;
{
	struct fet_chain_data	*cd;
	struct fet_list_data	*ld;
	int			i;

	cd = &sc->fet_cdata;
	ld = sc->fet_ldata;
	for (i = 0; i < FET_TX_LIST_CNT; i++) {
		cd->fet_tx_chain[i].fet_ptr = &ld->fet_tx_list[i];
		if (i == (FET_TX_LIST_CNT - 1))
			cd->fet_tx_chain[i].fet_nextdesc = 
				&cd->fet_tx_chain[0];
		else
			cd->fet_tx_chain[i].fet_nextdesc =
				&cd->fet_tx_chain[i + 1];
	}

	cd->fet_tx_free = &cd->fet_tx_chain[0];
	cd->fet_tx_tail = cd->fet_tx_head = NULL;

	return(0);
}


/*
 * Initialize the RX descriptors and allocate mbufs for them. Note that
 * we arrange the descriptors in a closed ring, so that the last descriptor
 * points back to the first.
 */
static int fet_list_rx_init(sc)
	struct fet_softc		*sc;
{
	struct fet_chain_data	*cd;
	struct fet_list_data	*ld;
	int			i;

	cd = &sc->fet_cdata;
	ld = sc->fet_ldata;

	for (i = 0; i < FET_RX_LIST_CNT; i++) {
		cd->fet_rx_chain[i].fet_ptr =
			(struct fet_desc *)&ld->fet_rx_list[i];
		if (fet_newbuf(sc, &cd->fet_rx_chain[i]) == ENOBUFS)
			return(ENOBUFS);
		if (i == (FET_RX_LIST_CNT - 1)) {
			cd->fet_rx_chain[i].fet_nextdesc =
					&cd->fet_rx_chain[0];
			ld->fet_rx_list[i].fet_next =
					vtophys(&ld->fet_rx_list[0]);
		} else {
			cd->fet_rx_chain[i].fet_nextdesc =
					&cd->fet_rx_chain[i + 1];
			ld->fet_rx_list[i].fet_next =
					vtophys(&ld->fet_rx_list[i + 1]);
		}
	}

	cd->fet_rx_head = &cd->fet_rx_chain[0];

	return(0);
}

/*
 * Initialize an RX descriptor and attach an MBUF cluster. 
 */
static int fet_newbuf(sc, c)
	struct fet_softc		*sc;
	struct fet_chain_onefrag	*c;
{
	struct mbuf		*m_new = NULL;

	MGETHDR(m_new, M_DONTWAIT, MT_DATA);
	if (m_new == NULL) {
		printf("fet%d: no memory for rx list -- packet dropped!\n",
								sc->fet_unit);
		return(ENOBUFS);
	}

	MCLGET(m_new, M_DONTWAIT);
	if (!(m_new->m_flags & M_EXT)) {
		printf("fet%d: no memory for rx list -- packet dropped!\n",
								sc->fet_unit);
		m_freem(m_new);
		return(ENOBUFS);
	}

	c->fet_mbuf = m_new;
	c->fet_ptr->fet_status = FET_RXSTAT;
	c->fet_ptr->fet_data = vtophys(mtod(m_new, caddr_t));
	c->fet_ptr->fet_ctl = FET_RXCTL | FET_RXLEN;

	return(0);
}

/*
 * A frame has been uploaded: pass the resulting mbuf chain up to
 * the higher level protocols.
 */
static void fet_rxeof(sc)
	struct fet_softc		*sc;
{
        struct ether_header	*eh;
        struct mbuf		*m;
        struct ifnet		*ifp;
	struct fet_chain_onefrag	*cur_rx;
	int			total_len = 0;
	u_int32_t		rxstat;

	ifp = &sc->arpcom.ac_if;

	while(!((rxstat = sc->fet_cdata.fet_rx_head->fet_ptr->fet_status) &
							FET_RXSTAT_OWN)) {
		cur_rx = sc->fet_cdata.fet_rx_head;
		sc->fet_cdata.fet_rx_head = cur_rx->fet_nextdesc;
		total_len = FET_RXBYTES(cur_rx->fet_ptr->fet_status);

                /* If the length of packet is larger than 1518, discard this packet */
                if(total_len> (MAX_PACKET_LEN+U_CRC_LEN) )
                   continue;
		/*
		 * If an error occurs, update stats, clear the
		 * status word and leave the mbuf cluster in place:
		 * it should simply get re-used next time this descriptor
	 	 * comes up in the ring.
		 */
		if (rxstat & FET_RXSTAT_RXERR) {
			ifp->if_ierrors++;
			printf("fet%d: rx error: ", sc->fet_unit);
			switch(rxstat & 0x000000FF) {
			case FET_RXSTAT_CRCERR:
				printf("crc error\n");
				break;
			case FET_RXSTAT_FRAMEALIGNERR:
				printf("frame alignment error\n");
				break;
			case FET_RXSTAT_FIFOOFLOW:
				printf("FIFO overflow\n");
				break;
			case FET_RXSTAT_GIANT:
				printf("received giant packet\n");
				break;
			case FET_RXSTAT_RUNT:
				printf("received runt packet\n");
				break;
			case FET_RXSTAT_BUSERR:
				printf("system bus error\n");
				break;
			case FET_RXSTAT_BUFFERR:
				printf("rx buffer error\n");
				break;
			default:
				printf("unknown rx error\n");
				break;
			}
			cur_rx->fet_ptr->fet_status = FET_RXSTAT;
			cur_rx->fet_ptr->fet_ctl = FET_RXCTL|FET_RXLEN;
			continue;
		}

		/* No errors; receive the packet. */	
		m = cur_rx->fet_mbuf;
		/* substract 4 bytes CRC */		
		total_len -= ETHER_CRC_LEN;

		/*
		 * Try to conjure up a new mbuf cluster. If that
		 * fails, it means we have an out of memory condition and
		 * should leave the buffer in place and continue. This will
		 * result in a lost packet, but there's little else we
		 * can do in this situation.
		 */
		if (fet_newbuf(sc, cur_rx) == ENOBUFS) {
			ifp->if_ierrors++;
			cur_rx->fet_ptr->fet_status = FET_RXSTAT;
			cur_rx->fet_ptr->fet_ctl = FET_RXCTL|FET_RXLEN;
			continue;
		}
                
		ifp->if_ipackets++;
		eh = mtod(m, struct ether_header *);
		m->m_pkthdr.rcvif = ifp;
		m->m_pkthdr.len = m->m_len = total_len;
#if NBPFILTER > 0
		/*
		 * Handle BPF listeners. Let the BPF user see the packet, but
		 * don't pass it up to the ether_input() layer unless it's
		 * a broadcast packet, multicast packet, matches our ethernet
		 * address or the interface is in promiscuous mode.
		 */
		if (ifp->if_bpf) {
			bpf_mtap(ifp, m);
			if (ifp->if_flags & IFF_PROMISC &&
				(bcmp(eh->ether_dhost, sc->arpcom.ac_enaddr,
						ETHER_ADDR_LEN) &&
					(eh->ether_dhost[0] & 1) == 0)) {
				m_freem(m);
				continue;
			}
		}
#endif
		/* Remove header from mbuf and pass it on. */
		m_adj(m, sizeof(struct ether_header));
		ether_input(ifp, eh, m);
	}

	return;
}

void fet_rxeoc(sc)
	struct fet_softc		*sc;
{

	fet_rxeof(sc);
	
	if (!fet_safe_rx_off(sc))
		printf("fet%d: RX shutdown error! \n", sc->fet_unit);
			
	CSR_WRITE_4(sc, FET_RXADDR, vtophys(sc->fet_cdata.fet_rx_head->fet_ptr));
	FET_SETBIT16(sc, FET_COMMAND, FET_CMD_RX_ON);	

	return;
}

/*
 * A frame was downloaded to the chip. It's safe for us to clean up
 * the list buffers.
 */

static void fet_txeof(sc)
	struct fet_softc		*sc;
{
	struct fet_chain		*cur_tx;
	struct ifnet		*ifp;
	register struct mbuf	*n;

	ifp = &sc->arpcom.ac_if;

	/* Clear the timeout timer. */
	ifp->if_timer = 0;

	/* Sanity check. */
	if (sc->fet_cdata.fet_tx_head == NULL)
		return;

	/*
	 * Go through our tx list and free mbufs for those
	 * frames that have been transmitted.
	 */
	while(sc->fet_cdata.fet_tx_head->fet_mbuf != NULL) {
		u_int32_t		txstat;

		cur_tx = sc->fet_cdata.fet_tx_head;
		
		if (cur_tx->fet_ptr->fet_ctl & FET_TXCTL_FIRSTFRAG) {
			txstat = cur_tx->fet_ptr->fet_status;

			if (txstat & FET_TXSTAT_OWN)
				break;

			if (txstat & FET_TXSTAT_ERRSUM) {
				ifp->if_oerrors++;
				if (txstat & FET_TXSTAT_DEFER)
					ifp->if_collisions++;
				if (txstat & FET_TXSTAT_LATECOLL)
					ifp->if_collisions++;
				}

			ifp->if_collisions +=(txstat & FET_TXSTAT_COLLCNT) >> 3;

			ifp->if_opackets++;
			}

        MFREE(cur_tx->fet_mbuf, n);
		cur_tx->fet_mbuf = NULL;

		if (sc->fet_cdata.fet_tx_head == sc->fet_cdata.fet_tx_tail) {
			sc->fet_cdata.fet_tx_head = NULL;
			sc->fet_cdata.fet_tx_tail = NULL;
			break;
		}

		sc->fet_cdata.fet_tx_head = cur_tx->fet_nextdesc;
	}

	return;
}

/*
 * TX 'end of channel' interrupt handler.
 */
static void fet_txeoc(sc)
	struct fet_softc		*sc;
{
	struct ifnet		*ifp;

	ifp = &sc->arpcom.ac_if;

	ifp->if_timer = 0;

	if (sc->fet_cdata.fet_tx_head == NULL) {
		ifp->if_flags &= ~IFF_OACTIVE;
		sc->fet_cdata.fet_tx_tail = NULL;
		/*
		if (sc->fet_want_auto)
			fet_autoneg_mii(sc, FET_FLAG_SCHEDDELAY, 1);
		*/
	}

	return;
}

static void fet_intr(arg)
	void			*arg;
{
	struct fet_softc		*sc;
	struct ifnet		*ifp;
	u_int16_t		status;

	sc = arg;
	ifp = &sc->arpcom.ac_if;

	/* Supress unwanted interrupts. */
	if (!(ifp->if_flags & IFF_UP)) {
		fet_stop(sc);
		return;
	}

	/* Disable interrupts. */
	CSR_WRITE_2(sc, FET_IMR, 0x0000);

	for (;;) {

		status = CSR_READ_2(sc, FET_ISR);
		if (status)
			CSR_WRITE_2(sc, FET_ISR, status);

		if ((status & IMRShadow) == 0)
			break;

		if (status & FET_ISR_PRX)
			fet_rxeof(sc);

		if ((status & FET_ISR_RXE) || (status & FET_ISR_RU) ||
		    (status & FET_ISR_RU) || (status & FET_ISR_OVFI) ||
		    (status & FET_ISR_PKTRACE)) {
			fet_rxeof(sc);
			fet_rxeoc(sc);
		}

		if (status & FET_ISR_PTX) {
			fet_txeof(sc);
			fet_txeoc(sc);
		}

		if ((status & FET_ISR_TU)||(status & FET_ISR_TXE)){ 
			ifp->if_oerrors++;
			fet_txeof(sc);
			if (sc->fet_cdata.fet_tx_head != NULL) {
				FET_SETBIT16(sc, FET_COMMAND, FET_CMD_TX_ON);
				FET_SETBIT16(sc, FET_COMMAND, FET_CMD_TDMD1);
			}
		}
		
		if (status & FET_ISR_SRCI)
			fet_link_change(sc);

		if (status & FET_ISR_BE) {
			fet_reset(sc);
			fet_init(sc);
		}
	}

	/* Re-enable interrupts. */
	CSR_WRITE_2(sc, FET_IMR, IMRShadow);

	if (ifp->if_snd.ifq_head != NULL) {
		fet_start(ifp);
	}

	return;
}

/*
 * Encapsulate an mbuf chain in a descriptor by coupling the mbuf data
 * pointers to the fragment pointers.
 */
static int fet_encap(sc, c, m_head)
	struct fet_softc		*sc;
	struct fet_chain		*c;
	struct mbuf		*m_head;
{
	int			frag = 0;
	struct fet_desc		*f = NULL;
	int			total_len;
	struct mbuf		*m;

	m = m_head;
	total_len = 0;

	/*
	 * The VIA Rhine wants packet buffers to be longword
	 * aligned, but very often our mbufs aren't. Rather than
	 * waste time trying to decide when to copy and when not
	 * to copy, just do it all the time.
	 */
	if (m != NULL) {
		struct mbuf		*m_new = NULL;

		MGETHDR(m_new, M_DONTWAIT, MT_DATA);
		if (m_new == NULL) {
			printf("fet%d: no memory for tx list", sc->fet_unit);
			return(1);
		}
		if (m_head->m_pkthdr.len > MHLEN) {
			MCLGET(m_new, M_DONTWAIT);
			if (!(m_new->m_flags & M_EXT)) {
				m_freem(m_new);
				printf("fet%d: no memory for tx list",
						sc->fet_unit);
				return(1);
			}
		}
		m_copydata(m_head, 0, m_head->m_pkthdr.len,	
					mtod(m_new, caddr_t));
		m_new->m_pkthdr.len = m_new->m_len = m_head->m_pkthdr.len;
		m_freem(m_head);
		m_head = m_new;
		/*
		 * The Rhine chip doesn't auto-pad, so we have to make
		 * sure to pad short frames out to the minimum frame length
		 * ourselves.
		 */
		if (m_head->m_len < FET_MIN_FRAMELEN) {
			m_new->m_pkthdr.len += FET_MIN_FRAMELEN - m_new->m_len;
			m_new->m_len = m_new->m_pkthdr.len;
		}
		f = c->fet_ptr;
		f->fet_data = vtophys(mtod(m_new, caddr_t));
		f->fet_ctl = total_len = m_new->m_len;
		f->fet_ctl |= FET_TXCTL_TLINK|FET_TXCTL_FIRSTFRAG;
		f->fet_status = 0;
		frag = 1;
	}

	c->fet_mbuf = m_head;
	c->fet_ptr->fet_ctl |= FET_TXCTL_LASTFRAG|FET_TXCTL_FINT;
	c->fet_ptr->fet_next = vtophys(c->fet_nextdesc->fet_ptr);

	return(0);
}

/*
 * Main transmit routine. To avoid having to do mbuf copies, we put pointers
 * to the mbuf data regions directly in the transmit lists. We also save a
 * copy of the pointers since the transmit list fragment pointers are
 * physical addresses.
 */

static void fet_start(ifp)
	struct ifnet		*ifp;
{
	struct fet_softc		*sc;
	struct mbuf		*m_head = NULL, *m = NULL;
	struct fet_chain		*cur_tx = NULL, *start_tx, *first_tx = NULL;
	int		frag_count;


	sc = ifp->if_softc;

/*
	if (sc->fet_autoneg) {
		sc->fet_tx_pend = 1;
		return;
	}
*/

	/*
	 * Check for an available queue slot. If there are none,
	 * punt.
	 */
	if (sc->fet_cdata.fet_tx_free->fet_mbuf != NULL) {
		ifp->if_flags |= IFF_OACTIVE;
		return;
	}

	start_tx = sc->fet_cdata.fet_tx_free;

	while(sc->fet_cdata.fet_tx_free->fet_mbuf == NULL) {

		IF_DEQUEUE(&ifp->if_snd, m_head);
		if (m_head == NULL)
			break;
		
		if (sc->fet_chip_revid < REV_ID_VT3065_A) {

			/* Pick a descriptor off the free list. */
			cur_tx = sc->fet_cdata.fet_tx_free;
			sc->fet_cdata.fet_tx_free = cur_tx->fet_nextdesc;
			/* Pack the data into the descriptor. */
			fet_encap(sc, cur_tx, m_head);

		if (cur_tx != start_tx)
			FET_TXOWN(cur_tx) = FET_TXSTAT_OWN;

#if NBPFILTER > 0
		/*
		 * If there's a BPF listener, bounce a copy of this frame
		 * to him.
		 */
		if (ifp->if_bpf)		
			bpf_mtap(ifp, cur_tx->fet_mbuf);
#endif

		FET_TXOWN(cur_tx) = FET_TXSTAT_OWN;

			}

		else {
			frag_count = 1;			 
			for (m = m_head; m != NULL; m = m->m_next) {

				/* Pick a descriptor off the free list. */
				cur_tx = sc->fet_cdata.fet_tx_free;
				sc->fet_cdata.fet_tx_free = cur_tx->fet_nextdesc;
				
				if (frag_count == 1)
					first_tx = cur_tx;
			
				cur_tx->fet_ptr->fet_data = vtophys(mtod(m, caddr_t));

				/* pad the short frame to minimum frame length */
				/* ps: Pad those packets which only occupy 1 fragment */
				if (m->m_len < FET_MIN_FRAMELEN && frag_count == 1 && m->m_next == NULL) {
					cur_tx->fet_ptr->fet_ctl = FET_MIN_FRAMELEN;
					m->m_pkthdr.len += FET_MIN_FRAMELEN - m->m_len;
					m->m_len = m->m_pkthdr.len;
					}
				else
					cur_tx->fet_ptr->fet_ctl = m->m_len;
				
				if (frag_count == 1)
					cur_tx->fet_ptr->fet_ctl |= FET_TXCTL_FINT|FET_TXCTL_TLINK|		\
													FET_TXCTL_FIRSTFRAG;
				else
					cur_tx->fet_ptr->fet_ctl |= FET_TXCTL_FINT|FET_TXCTL_TLINK;

				cur_tx->fet_ptr->fet_status = 0;
	
				cur_tx->fet_mbuf = m;			
				cur_tx->fet_ptr->fet_next = vtophys(cur_tx->fet_nextdesc->fet_ptr);
				
				frag_count++;				

#if NBPFILTER > 0
				/*
		 		* If there's a BPF listener, bounce a copy of this frame
		 		* to him.
		 		*/
				if (ifp->if_bpf)		
					bpf_mtap(ifp, cur_tx->fet_mbuf);
#endif
				} /* end for */
			
			cur_tx->fet_ptr->fet_ctl |=	FET_TXCTL_LASTFRAG;

			FET_TXOWN(first_tx) = FET_TXSTAT_OWN;
			
			} /* end else */
					
		FET_SETBIT16(sc, FET_COMMAND, FET_CMD_TX_ON|FET_CMD_TDMD1);
	}

	/*
	 * If there are no frames queued, bail.
	 */
	if (cur_tx == NULL)
		return;

	sc->fet_cdata.fet_tx_tail = cur_tx;

	if (sc->fet_cdata.fet_tx_head == NULL)
		sc->fet_cdata.fet_tx_head = start_tx;

	/*
	 * Set a timeout in case the chip goes out to lunch.
	 */
	ifp->if_timer = 5;

	return;
}

static void fet_init(xsc)
	void			*xsc;
{
	struct fet_softc		*sc = xsc;
	struct ifnet		*ifp = &sc->arpcom.ac_if;	
	int			s;
	u_int16_t	FDXFlag;
#ifdef DEBUG
	u_int8_t	aa;
	char		c[5];
#endif

	if (sc->fet_autoneg)
		return;

	s = splimp();
		
	/*
	 * Cancel pending I/O and free all RX/TX buffers.
	 */
	fet_stop(sc);
	
	/* if vt3065 */    
     if (sc->fet_chip_revid >= REV_ID_VT3065_A ) { 
         /* clear sticky bit before reset & read ethernet address  */      
         CSR_WRITE_1(sc, FET_STICKYHW, CSR_READ_1(sc, FET_STICKYHW) & 0xFC);              
         /* disable force PME-enable */
         CSR_WRITE_1(sc, FET_WOLCGCLR, 0x80);
         /* disable power-event config bit */
         CSR_WRITE_1(sc, FET_WOLCRCLR, 0xFF);
         /* clear power status */
         CSR_WRITE_1(sc, FET_PWRCSRCLR, 0xFF);   
    }
    
    fet_reset(sc);
    
    /* issue AUTOLoad in EECSR to reload eeprom */
    
    CSR_WRITE_1(sc, FET_MACREGCSR, 0x20);

    /* if vt3065 delay after reset */
    if (sc->fet_chip_revid >= REV_ID_VT3065_A ) { 
           post_reset_delay();
           /* for 3065D, EEPROM reloaded will cause bit 0 in MAC_REG_CFGA turned on.
            it makes MAC receive magic packet automatically. So, driver turn it off. */
           
           CSR_WRITE_1(sc, FET_CFGA,CSR_READ_1(sc, FET_CFGA) & 0xFE);
    }

	/* set the MIIAD again because software reset will reset the value */
     FET_SETBIT(sc, FET_MIIADDR, FET_MIIADDR_MSRCEN | FET_MIIADDR_MAD0);

	/* set TCR RCR threshold */
	/* set DMAL to 110 , and not to change REQOPT set in attach fucntion */
	FET_SETBIT(sc, FET_BCR0, FET_BCR0_DMAL2 | FET_BCR0_DMAL1);	
	CSR_WRITE_1(sc, FET_BCR1, 0x00);    

	/* Set RX threshhold */
	FET_CLRBIT(sc, FET_RXCFG, FET_RXCFG_RX_THRESH);
	FET_SETBIT(sc, FET_RXCFG, FET_RXTHRESH_1024BYTES);
	/* Set TX threshhold */
	FET_CLRBIT(sc, FET_TXCFG, FET_TXCFG_TX_THRESH);
	FET_SETBIT(sc, FET_TXCFG, FET_TXTHRESH_512BYTES);

	/* QPacket setting */
	if (sc->fet_chip_revid < REV_ID_VT3065_A)
            /* disable queue packet */
            FET_SETBIT(sc, FET_CFGB, FET_CFGB_QPKTDIS);
	else {		
            /* enable queue packet */
            FET_CLRBIT(sc, FET_CFGB, FET_CFGB_QPKTDIS);
        }

	/* set backoff algorithm to */
        FET_CLRBIT(sc, FET_CFGD,(FET_CFGD_CAP|FET_CFGD_CRADOM|FET_CFGD_MBA|FET_CFGD_BAKOPT));

	/* Init circular RX list. */
	if (fet_list_rx_init(sc) == ENOBUFS) {
		printf("fet%d: initialization failed: no "
			"memory for rx buffers\n", sc->fet_unit);
		fet_stop(sc);
		(void)splx(s);
		return;
	}

	/*
	 * Init tx descriptors.
	 */
	fet_list_tx_init(sc);

	/* If we want promiscuous mode, set the allframes bit. */
	if (ifp->if_flags & IFF_PROMISC)
		FET_SETBIT(sc, FET_RXCFG, FET_RXCFG_RX_PROMISC);
	else
		FET_CLRBIT(sc, FET_RXCFG, FET_RXCFG_RX_PROMISC);

	/* Set capture broadcast bit to capture broadcast frames. */
	if (ifp->if_flags & IFF_BROADCAST)
		FET_SETBIT(sc, FET_RXCFG, FET_RXCFG_RX_BROAD);
	else
		FET_CLRBIT(sc, FET_RXCFG, FET_RXCFG_RX_BROAD);

	/*
	 * Program the multicast filter, if necessary.
	 */
	fet_setmulti(sc);

	/*
	 * Load the address of the RX list.
	 */
	CSR_WRITE_4(sc, FET_RXADDR, vtophys(sc->fet_cdata.fet_rx_head->fet_ptr));		
	CSR_WRITE_4(sc, FET_TXADDR, vtophys(&sc->fet_ldata->fet_tx_list[0]));

	
    /* query fullduplex */
    FDXFlag = fet_query_auto(sc);

	if (FDXFlag == 1)		 	
		FET_SETBIT16(sc, FET_COMMAND, FET_CMD_FDX);       /* Set Chip Fullduplex mode */		
	else
		FET_CLRBIT16(sc, FET_COMMAND, FET_CMD_FDX);      /* Set Chip Halfduplex mode */		

#ifdef DEBUG
	aa = CSR_READ_1(sc, FET_MIICR);
	printf("MIICR: %x \n",aa);
	aa = CSR_READ_1(sc, FET_MIIADDR);
	printf("MIIADDR: %x \n",aa);
	db_readline(c,4);
#endif
	   
	/* Enable receiver and transmitter. */
	CSR_WRITE_2(sc, FET_COMMAND, FET_CMD_TX_NOPOLL|FET_CMD_START|
				    FET_CMD_TX_ON|FET_CMD_RX_ON);
				         
	/*
	 * Enable interrupts.
	 */	
	CSR_WRITE_2(sc, FET_IMR, IMRShadow);
	
	ifp->if_flags |= IFF_RUNNING;
	ifp->if_flags &= ~IFF_OACTIVE;

	(void)splx(s);

	return;
}

/*
 * Set media options.
 */
static int fet_ifmedia_upd(ifp)
	struct ifnet		*ifp;
{
	struct fet_softc		*sc;
	struct ifmedia		*ifm;

	sc = ifp->if_softc;
	ifm = &sc->ifmedia;

	if (IFM_TYPE(ifm->ifm_media) != IFM_ETHER)
		return(EINVAL);
	/*
	if (IFM_SUBTYPE(ifm->ifm_media) == IFM_AUTO)
		fet_autoneg_mii(sc, FET_FLAG_SCHEDDELAY, 1);
	else
		fet_setmode_mii(sc, ifm->ifm_media);
	*/

	return(0);
}

/*
 * Report current media status.
 */
static void fet_ifmedia_sts(ifp, ifmr)
	struct ifnet		*ifp;
	struct ifmediareq	*ifmr;
{
	struct fet_softc		*sc;
	u_int8_t	SRbak;
    u_int16_t	FDXFlag;
	/*
	u_int16_t		advert = 0, ability = 0;
	*/

	sc = ifp->if_softc;

	ifmr->ifm_active = IFM_ETHER;

	/* query fullduplex */
	FDXFlag = fet_query_auto(sc);

	/* query speed */
	SRbak = CSR_READ_1(sc, FET_MIISR);

	if(FDXFlag)
	   ifmr->ifm_active |= (SRbak & FET_MIISR_SPEED)? (IFM_10_T & IFM_FDX): (IFM_100_TX & IFM_FDX);
	else
	   ifmr->ifm_active |= (SRbak & FET_MIISR_SPEED)? (IFM_10_T & IFM_HDX): (IFM_100_TX & IFM_HDX);
	
	/*
	if (!(fet_phy_readreg(sc, PHY_BMCR) & PHY_BMCR_AUTONEGENBL)) {
		if (fet_phy_readreg(sc, PHY_BMCR) & PHY_BMCR_SPEEDSEL)
			ifmr->ifm_active = IFM_ETHER|IFM_100_TX;
		else
			ifmr->ifm_active = IFM_ETHER|IFM_10_T;
		if (fet_phy_readreg(sc, PHY_BMCR) & PHY_BMCR_DUPLEX)
			ifmr->ifm_active |= IFM_FDX;
		else
			ifmr->ifm_active |= IFM_HDX;
		return;
	}

	ability = fet_phy_readreg(sc, PHY_LPAR);
	advert = fet_phy_readreg(sc, PHY_ANAR);
	if (advert & PHY_ANAR_100BT4 &&
		ability & PHY_ANAR_100BT4) {
		ifmr->ifm_active = IFM_ETHER|IFM_100_T4;
	} else if (advert & PHY_ANAR_100BTXFULL &&
		ability & PHY_ANAR_100BTXFULL) {
		ifmr->ifm_active = IFM_ETHER|IFM_100_TX|IFM_FDX;
	} else if (advert & PHY_ANAR_100BTXHALF &&
		ability & PHY_ANAR_100BTXHALF) {
		ifmr->ifm_active = IFM_ETHER|IFM_100_TX|IFM_HDX;
	} else if (advert & PHY_ANAR_10BTFULL &&
		ability & PHY_ANAR_10BTFULL) {
		ifmr->ifm_active = IFM_ETHER|IFM_10_T|IFM_FDX;
	} else if (advert & PHY_ANAR_10BTHALF &&
		ability & PHY_ANAR_10BTHALF) {
		ifmr->ifm_active = IFM_ETHER|IFM_10_T|IFM_HDX;
	}
	*/

	return;
}

static int fet_ioctl(ifp, command, data)
	struct ifnet		*ifp;
	u_long			command;
	caddr_t			data;
{
	struct fet_softc		*sc = ifp->if_softc;
	struct ifreq		*ifr = (struct ifreq *) data;
	int			s, error = 0;

	s = splimp();

	switch(command) {
	case SIOCSIFADDR:
	case SIOCGIFADDR:
	case SIOCSIFMTU:
		error = ether_ioctl(ifp, command, data);
		break;
	case SIOCSIFFLAGS:
		if (ifp->if_flags & IFF_UP) {
			fet_init(sc);
		} else {
			if (ifp->if_flags & IFF_RUNNING)
				fet_stop(sc);
		}
		error = 0;
		break;
	case SIOCADDMULTI:
	case SIOCDELMULTI:
		fet_setmulti(sc);
		error = 0;
		break;
	case SIOCGIFMEDIA:
	case SIOCSIFMEDIA:
		error = ifmedia_ioctl(ifp, ifr, &sc->ifmedia, command);
		break;
	default:
		error = EINVAL;
		break;
	}

	(void)splx(s);

	return(error);
}

static void fet_watchdog(ifp)
	struct ifnet		*ifp;
{
	struct fet_softc		*sc;

	sc = ifp->if_softc;
	
	/*
	if (sc->fet_autoneg) {
		fet_autoneg_mii(sc, FET_FLAG_DELAYTIMEO, 1);
		return;
	}
	*/

	ifp->if_oerrors++;
	printf("fet%d: watchdog timeout\n", sc->fet_unit);

	/*
	if (!(fet_phy_readreg(sc, PHY_BMSR) & PHY_BMSR_LINKSTAT))
		printf("fet%d: no carrier - transceiver cable problem?\n",
								sc->fet_unit);
	*/

	fet_stop(sc);
	fet_reset(sc);
	fet_init(sc);

	if (ifp->if_snd.ifq_head != NULL)
		fet_start(ifp);

	return;
}

/*
 * Stop the adapter and free any mbufs allocated to the
 * RX and TX lists.
 */
static void fet_stop(sc)
	struct fet_softc		*sc;
{
	register int		i;
	struct ifnet		*ifp;

	ifp = &sc->arpcom.ac_if;
	ifp->if_timer = 0;
	
	FET_CLRBIT16(sc, FET_COMMAND, FET_CMD_TX_ON);

	if (!fet_safe_rx_off(sc))
		printf("fet%d: RX shutdown error! \n", sc->fet_unit);

	FET_SETBIT16(sc, FET_COMMAND, FET_CMD_STOP);

	CSR_WRITE_2(sc, FET_IMR, 0x0000);
	CSR_WRITE_4(sc, FET_TXADDR, 0x00000000);
	CSR_WRITE_4(sc, FET_RXADDR, 0x00000000);

	/*
	 * Free data in the RX lists.
	 */
	for (i = 0; i < FET_RX_LIST_CNT; i++) {
		if (sc->fet_cdata.fet_rx_chain[i].fet_mbuf != NULL) {
			m_freem(sc->fet_cdata.fet_rx_chain[i].fet_mbuf);
			sc->fet_cdata.fet_rx_chain[i].fet_mbuf = NULL;
		}
	}
	bzero((char *)&sc->fet_ldata->fet_rx_list,
		sizeof(sc->fet_ldata->fet_rx_list));

	/*
	 * Free the TX list buffers.
	 */
	for (i = 0; i < FET_TX_LIST_CNT; i++) {
		if (sc->fet_cdata.fet_tx_chain[i].fet_mbuf != NULL) {
			m_freem(sc->fet_cdata.fet_tx_chain[i].fet_mbuf);
			sc->fet_cdata.fet_tx_chain[i].fet_mbuf = NULL;
		}
	}

	bzero((char *)&sc->fet_ldata->fet_tx_list,
		sizeof(sc->fet_ldata->fet_tx_list));

	ifp->if_flags &= ~(IFF_RUNNING | IFF_OACTIVE);

	return;
}

/*
 * Stop all chip I/O so that the kernel's probe routines don't
 * get confused by errant DMAs when rebooting.
 */
static void fet_shutdown(howto, arg)
	int			howto;
	void			*arg;
{
	struct fet_softc		*sc = (struct fet_softc *)arg;

	fet_stop(sc);

	return;
}

static struct pci_device fet_device = {
	"fet",
	fet_probe,
	fet_attach,
	&fet_count,
	NULL
};
DATA_SET(pcidevice_set, fet_device);
